"""
Organizations API Module Test Cases
"""
import itertools
from unittest.mock import patch

import ddt
from django.test import override_settings
from opaque_keys.edx.keys import CourseKey

from organizations import api
from organizations import exceptions
from organizations import models
from organizations.tests import utils
from organizations.data import log as data_module_logger


@ddt.ddt
class OrganizationsApiTestCase(utils.OrganizationsTestCaseBase):
    """
    Main Test Case module for Organizations API
    """
    def setUp(self):
        """
        Organizations API Test Case scaffolding
        """
        super().setUp()
        self.test_organization = api.add_organization({
            'short_name': 'test_organization',
            'name': 'test_organizationßßß',
            'description': 'Test Organization Descriptionßßß'
        })

    def test_add_organization(self):
        """ Unit Test: test_add_organization"""
        with self.assertNumQueries(3):
            organization = api.add_organization({
                'short_name': 'local_organization',
                'name': 'local_organizationßßß',
                'description': 'Local Organization Descriptionßßß'
            })
        self.assertGreater(organization['id'], 0)

    def test_add_organization_active_exists(self):
        """ Unit Test: test_add_organization_active_exists"""
        organization_data = {
            'short_name': 'local_organization',
            'name': 'local_organizationßßß',
            'description': 'Local Organization Descriptionßßß'
        }
        organization = api.add_organization(organization_data)
        self.assertGreater(organization['id'], 0)
        with self.assertNumQueries(1):
            organization = api.add_organization(organization_data)

    def test_add_organization_inactive_to_active(self):
        """ Unit Test: test_add_organization_inactive_to_active"""
        organization_data = {
            'short_name': 'local_organization',
            'name': 'local_organizationßßß',
            'description': 'Local Organization Descriptionßßß'
        }
        organization = api.add_organization(organization_data)
        self.assertGreater(organization['id'], 0)
        api.remove_organization(organization['id'])

        with self.assertNumQueries(5):
            organization = api.add_organization(organization_data)

        # Assert that the organization is active by making sure we can load it.
        loaded_organization = api.get_organization(organization['id'])
        self.assertEqual(loaded_organization['name'], organization_data['name'])

    def test_add_organization_inactive_organization_with_relationships(self):
        """ Unit Test: test_add_organization_inactive_organization_with_relationships"""
        organization_data = {
            'short_name': 'local_organization',
            'name': 'local_organizationßßß',
            'description': 'Local Organization Descriptionßßß'
        }
        organization = api.add_organization(organization_data)
        api.add_organization_course(
            organization,
            self.test_course_key
        )

        with self.assertNumQueries(1):
            organization = api.add_organization(organization_data)

    @ddt.data(
        # Empty short name
        {
            'short_name': '',
            'name': 'local_organizationßßß',
            'description': 'Local Organization Description 2ßßß'
        },
        # Empty name
        {
            'short_name': 'local_organization',
            'name': '',
            'description': 'Local Organization Description 2ßßß'
        },
        # Missing short name
        {
            'name': 'local_organizationßßß',
            'description': 'Local Organization Description 2ßßß'
        },
        # Missing name
        {
            'short_name': 'local_organization',
            'description': 'Local Organization Description 2ßßß'
        },
    )
    def test_add_organization_invalid_data_throws_exceptions(self, organization_data):
        """ Unit Test: test_add_organization_invalid_data_throws_exceptions"""
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.add_organization(organization_data)

    def test_edit_organization(self):
        """ Unit Test: test_edit_organization"""
        self.test_organization['name'] = 'Edited Organizationßßß'

        with self.assertNumQueries(1):
            api.edit_organization(self.test_organization)

    def test_edit_organization_invalid_data_throws_exceptions(self):
        """ Unit Test: test_edit_organization_invalid_data_throws_exceptions """
        self.test_organization['name'] = ''
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.edit_organization(self.test_organization)

        self.test_organization['id'] = 0
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.edit_organization(self.test_organization)

    def test_edit_organization_bogus_organization(self):
        """ Unit Test: test_edit_organization_bogus_organization """
        self.test_organization['id'] = 12345
        self.test_organization['name'] = 'bogus.organizationsßßß'
        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.edit_organization(self.test_organization)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.edit_organization(None)

    def test_get_organization(self):
        """ Unit Test: test_get_organization"""
        with self.assertNumQueries(1):
            organization = api.get_organization(self.test_organization['id'])
        self.assertEqual(organization['name'], self.test_organization['name'])
        self.assertEqual(organization['description'], self.test_organization['description'])

    def test_get_organization_by_short_name(self):
        """ Unit Test: get_organization_by_short_name"""
        api.add_organization({
            'name': 'local_organization_1ßßß',
            'short_name': 'Orgx1',
            'description': 'Local Organization 1 Descriptionßßß'
        })
        api.add_organization({
            'name': 'local_organization_2',
            'short_name': 'Orgx2',
            'description': 'Local Organization 2'
        })
        with self.assertNumQueries(1):
            organization = api.get_organization_by_short_name('Orgx2')
        self.assertEqual(organization['name'], 'local_organization_2')
        self.assertEqual(organization['description'], 'Local Organization 2')

        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.get_organization_by_short_name(None)

        with self.assertNumQueries(1):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.get_organization_by_short_name('not_existing')

    def test_get_organizations(self):
        """ Unit Test: test_get_organizations """
        api.add_organization({
            'name': 'local_organization_1ßßß',
            'short_name': 'Orgx1',
            'description': 'Local Organization 1 Descriptionßßß'
        })
        api.add_organization({
            'name': 'local_organization_2ßßß',
            'short_name': 'Orgx2',
            'description': 'Local Organization 2 Descriptionßßß'
        })
        with self.assertNumQueries(1):
            organizations = api.get_organizations()
        self.assertEqual(len(organizations), 3)  # One from SetUp, two from local

    def test_get_organization_invalid_organization(self):
        """ Unit Test: test_get_organization_invalid_organization """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.get_organization(None)
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidOrganizationException):
                api.get_organization(0)

    def test_remove_organization(self):
        """ Unit Test: test_remove_organization """
        with self.assertNumQueries(4):
            api.remove_organization(self.test_organization['id'])
        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.get_organization(self.test_organization['id'])

    def test_remove_organization_bogus_organization(self):
        """ Unit Test: test_remove_organization_bogus_organization """
        with self.assertNumQueries(4):
            api.remove_organization(self.test_organization['id'])

        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.get_organization(self.test_organization['id'])

        # Do it again with the valid id to hit the exception workflow
        with self.assertNumQueries(2):
            api.remove_organization(self.test_organization['id'])

        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.get_organization(self.test_organization['id'])

    def test_add_organization_course(self):
        """ Unit Test: test_add_organization_course """
        with self.assertNumQueries(3):
            api.add_organization_course(
                self.test_organization,
                self.test_course_key
            )

    def test_add_organization_course_active_exists(self):
        """ Unit Test: test_add_organization_course_active_exists """
        api.add_organization_course(
            self.test_organization,
            self.test_course_key
        )
        with self.assertNumQueries(1):
            api.add_organization_course(
                self.test_organization,
                self.test_course_key
            )

    def test_add_organization_course_inactive_to_active(self):
        """ Unit Test: test_add_organization_course_inactive_to_active """
        api.add_organization_course(
            self.test_organization,
            self.test_course_key
        )
        api.remove_organization_course(self.test_organization, self.test_course_key)
        with self.assertNumQueries(4):
            api.add_organization_course(
                self.test_organization,
                self.test_course_key
            )

    def test_add_organization_course_bogus_course_key(self):
        """ Unit Test: test_add_organization_course_bogus_course_key """
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidCourseKeyException):
                api.add_organization_course(self.test_organization, '12345667avßßß')
        with self.assertNumQueries(0):
            with self.assertRaises(exceptions.InvalidCourseKeyException):
                api.add_organization_course(self.test_organization, None)

    def test_get_course_organizations(self):
        """ Unit Test: test_get_course_organizations """
        api.add_organization_course(
            self.test_organization,
            self.test_course_key
        )
        with self.assertNumQueries(1):
            course_organizations = api.get_course_organizations(self.test_course_key)
        self.assertEqual(len(course_organizations), 1)

    def test_get_course_organization_no_linked_orgs(self):
        """
        Test that when a course is linked to no organizations,
        ``get_course_organization`` and ``get_course_organization_id`` return None.
        """
        assert api.get_course_organization(self.test_course_key) is None
        assert api.get_course_organization_id(self.test_course_key) is None

    def test_get_course_organization_multi_linked_orgs(self):
        """
        Test that when a course is linked to multiple organizations,
        ``get_course_organization`` and ``get_course_organization_id``
        return the first-linked one.
        """
        # Use non-alphabetically-ordered org names to test that the
        # returned org was the first *linked*, not just the first *alphabetically*.
        api.add_organization_course(
            api.add_organization({'short_name': 'orgW', 'name': 'Org West'}),
            self.test_course_key,
        )
        api.add_organization_course(
            api.add_organization({'short_name': 'orgN', 'name': 'Org North'}),
            self.test_course_key,
        )
        api.add_organization_course(
            api.add_organization({'short_name': 'orgS', 'name': 'Org South'}),
            self.test_course_key,
        )
        org_result = api.get_course_organization(self.test_course_key)
        assert org_result['short_name'] == 'orgW'
        org_id_result = api.get_course_organization_id(self.test_course_key)
        assert org_id_result
        assert org_id_result == org_result['id']

    def test_remove_organization_course(self):
        """ Unit Test: test_remove_organization_course """
        api.add_organization_course(
            self.test_organization,
            self.test_course_key
        )
        organizations = api.get_course_organizations(self.test_course_key)
        self.assertEqual(len(organizations), 1)
        with self.assertNumQueries(4):
            api.remove_organization_course(self.test_organization, self.test_course_key)
        organizations = api.get_course_organizations(self.test_course_key)
        self.assertEqual(len(organizations), 0)

    def test_remove_organization_course_missing_organization(self):
        """ Unit Test: test_remove_organization_course_missing_organization """
        with self.assertNumQueries(1):
            api.remove_organization_course(self.test_organization, self.test_course_key)
        organizations = api.get_course_organizations(self.test_course_key)
        self.assertEqual(len(organizations), 0)

    def test_remove_organization_course_missing_course(self):
        """ Unit Test: test_remove_organization_course_missing_organization """
        api.add_organization_course(
            self.test_organization,
            'edX/DemoX/Demo_Course'
        )
        organizations = api.get_course_organizations('edX/DemoX/Demo_Course')
        self.assertEqual(len(organizations), 1)
        with self.assertNumQueries(1):
            api.remove_organization_course(self.test_organization, self.test_course_key)
        organizations = api.get_course_organizations(self.test_course_key)
        self.assertEqual(len(organizations), 0)

    def test_remove_course_references(self):
        """ Unit Test: test_remove_course_references """
        # Add a course dependency on the test organization
        api.add_organization_course(
            self.test_organization,
            self.test_course_key
        )
        self.assertEqual(len(api.get_organization_courses(self.test_organization)), 1)

        # Remove the course dependency
        with self.assertNumQueries(3):
            api.remove_course_references(self.test_course_key)
        self.assertEqual(len(api.get_organization_courses(self.test_organization)), 0)

    @patch.object(api.log, 'info')
    def test_ensure_organization_retrieves_known_org(self, mock_log_info):
        """
        Test that, with oranization auto-create enabled, ``ensure_organization``
        returns the org data of an existing organization.
        """
        api.add_organization({
            'short_name': 'myorg', 'name': 'My Org', 'description': 'this is my org'
        })
        org_data = api.ensure_organization('myorg')
        assert org_data['id']
        assert org_data['short_name'] == 'myorg'
        assert org_data['name'] == 'My Org'
        assert org_data['description'] == 'this is my org'

        # We expect a single logging message in test_ensure_organization_creates_unknown_org,
        # so check for 0 log messages here as a control.
        assert mock_log_info.call_count == 0

    @patch.object(api.log, 'info')
    def test_ensure_organization_creates_unknown_org(self, mock_log_info):
        """
        Test that, with organization auto-create enabled, ``ensure_organization``
        automatically creates not-yet-existent organizations.
        """
        org_data = api.ensure_organization('myorg')
        assert org_data['id']
        assert org_data['short_name'] == 'myorg'
        assert org_data['name'] == 'myorg'
        assert org_data['description'] == ''  # auto-created, so no description.

        # Make sure we logged about the new organization's creation.
        assert mock_log_info.call_count == 1

        # Make sure that the organization has, in fact, been saved to the database
        # by loading it up again.
        assert api.get_organization_by_short_name('myorg') == org_data

    @override_settings(ORGANIZATIONS_AUTOCREATE=False)
    def test_ensure_organization_retrieves_known_org_no_autocreate(self):
        """
        Test that, with organization auto-create DISABLED, ``ensure_organization``
        returns the org data of an existing organization
        (just like it does when organization auto-create is enabled).
        """
        api.add_organization({
            'short_name': 'myorg', 'name': 'My Org', 'description': 'this is my org'
        })
        org_data = api.ensure_organization('myorg')
        assert org_data['id']
        assert org_data['short_name'] == 'myorg'
        assert org_data['name'] == 'My Org'
        assert org_data['description'] == 'this is my org'

    @override_settings(ORGANIZATIONS_AUTOCREATE=False)
    def test_ensure_organization_raises_for_unknown_org_no_autocreate(self):
        """
        Test that, with organization auto-create DISABLED, ``ensure_organization``
        fails when given a non-existent organization
        (in contrast to how it behaves when organization auto-create is enabled).
        """
        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.ensure_organization('myorg')

        # Make sure the organization was not saved to the database.
        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.get_organization_by_short_name('myorg')

    def test_autocreate_enabled_by_default(self):
        """
        Test that, by default, automatic organization creation is enabled.
        """
        assert api.is_autocreate_enabled()


class BulkAddOrganizationsTestCase(utils.OrganizationsTestCaseBase):
    """
    Tests for `api.bulk_add_organizations`.
    """

    def test_validation_errors(self):
        """
        Test the `bulk_add_organizations` raises validation errors on bad input,
        and no organizations are created.
        """
        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.bulk_add_organizations([
                self.make_organization_data("valid_org"),
                {"description": "org with no short_name!"},
            ])
        assert len(api.get_organizations()) == 0

    def test_add_no_organizations(self):
        """
        Test that `bulk_add_organizations` is a no-op when given an empty list.
        """
        api.bulk_add_organizations([])
        assert len(api.get_organizations()) == 0

    def test_dry_run(self):
        """
        Test that `bulk_add_organizations` does nothing when `dry_run` is
        specified (except logging).
        """
        would_create, would_reactivate = api.bulk_add_organizations(
            [self.make_organization_data("org_a")],
            dry_run=True,
        )

        assert api.get_organizations() == []
        assert would_create == {"org_a"}
        assert would_reactivate == set()

    @patch.object(data_module_logger, 'info', autospec=True)
    def test_edge_cases(self, mock_log_info):
        """
        Test that bulk_add_organizations handles a few edge cases as expected.
        """
        # Add three orgs, and remove all but the first.
        # Use capitalized name to confirm case insensitivity when checking
        # for existing orgs.
        api.add_organization(self.make_organization_data("EXISTING_ORG"))
        api.remove_organization(
            api.add_organization(
                self.make_organization_data("org_to_reactivate")
            )["id"]
        )
        api.remove_organization(
            api.add_organization(
                self.make_organization_data("org_to_leave_inactive")
            )["id"]
        )

        # 1 query to load list of existing orgs,
        # 1 query to filter for only inactive existing orgs,
        # 1 query for create, and 1 query for update.
        with self.assertNumQueries(4):
            created, reactivated = api.bulk_add_organizations([

                # New organization.
                self.make_organization_data("org_X"),

                # Modify existing active organization; should be no-op.
                {
                    **self.make_organization_data("existing_org"),
                    "description": "this name should be ignored"
                },

                # Deleted organizations are still stored in the DB as "inactive".
                # Bulk-adding should reactivate it.
                self.make_organization_data("org_to_reactivate"),

                # Another new organizaiton.
                self.make_organization_data("org_Y"),

                # Another org with same short name (case-insensitively)
                # as first new organization; should be ignored.
                {**self.make_organization_data("ORG_x"), "name": "this name should be ignored"}
            ])

        # There should exist the already-existing org, the org that existed as inactive
        # but is not activated, and the two new orgs.
        # This should not include `org_to_leave_inactive`.
        organizations = api.get_organizations()
        assert {
            organization["short_name"] for organization in organizations
        } == {
            "EXISTING_ORG", "org_to_reactivate", "org_X", "org_Y"
        }

        # Based on the return value, make sure the expected breakdown of
        #    created vs. reactivated vs. not touched
        # is correct for the organizations passed to `bulk_add_organizations`.
        assert reactivated == {"org_to_reactivate"}
        assert created == {"org_X", "org_Y"}

        # Organization dicts with already-taken short_names shouldn't have modified
        # the existing orgs.
        assert "this name should be ignored" not in {
            organization["name"] for organization in organizations
        }

        # Based on logging messages, make sure we dropped the appropriate
        # organization dict from the bulk-add batch.
        logging_of_drop_from_batch = mock_log_info.call_args_list[0][0]
        assert logging_of_drop_from_batch[1]["short_name"] == (
            # We dropped this org data:
            self.make_organization_data("ORG_x")["short_name"]
        )
        assert logging_of_drop_from_batch[2]["short_name"] == (
            # in favor of this org data, which came earlier in the batch.
            self.make_organization_data("org_X")["short_name"]
        )

    def test_creation_of_inactive_organizations(self):
        """
        Test that the `activate` parameter can be set to `False`, which should
        cause new organizations to be created as "inactive" and stop organizations
        from being re-activated.
        """
        # Create one active and one inactive organization.
        api.add_organization(self.make_organization_data("existing_active_org"))
        api.remove_organization(
            api.add_organization(
                self.make_organization_data("existing_inactive_org")
            )["id"]
        )

        # Bulk-add some organizations.
        organizations_to_bulk_add = [
            # New organization.
            # Should be created as "inactive".
            self.make_organization_data("new_org"),

            # Existing active organization (with different description and capitalized
            # short-name). This should have no effect; the org should remain active.
            {
                **self.make_organization_data("EXISTING_ACTIVE_ORG"),
                "description": "this should be ignored"
            },

            # Existing inactive organization (with different description and capitalized
            # short-name). This should have no effect; the org should remain inactive.
            {
                **self.make_organization_data("EXISTING_INACTIVE_ORG"),
                "description": "this too should be ignored"
            },
        ]
        created, reactivated = api.bulk_add_organizations(
            organizations_to_bulk_add, activate=False
        )

        # Based on the return value, make sure the expected breakdown of
        #    created vs. reactivated vs. not touched
        # is correct for the organizations passed to `bulk_add_organizations`.
        assert reactivated == set()
        assert created == {"new_org"}

        # Check that there are a total of three organizations (including inactive ones).
        assert models.Organization.objects.count() == 3

        # Check that the new organization was created as inactive.
        new_org = models.Organization.objects.get(short_name="new_org")
        assert not new_org.active

        # Check that the active organization is unmodified.
        existing_active_org = models.Organization.objects.get(short_name="existing_active_org")
        assert existing_active_org.active
        assert existing_active_org.description == "Description of existing_active_org"

        # Check that the inactive organization is unmodified.
        existing_inactive_org = models.Organization.objects.get(short_name="existing_inactive_org")
        assert not existing_inactive_org.active
        assert existing_inactive_org.description == "Description of existing_inactive_org"

    def test_add_several_organizations(self):
        """
        Test that the query_count of bulk_add_organizations does not increase
        when given more organizations.
        """
        existing_org = api.add_organization(self.make_organization_data("existing_org"))
        api.remove_organization(
            api.add_organization(
                self.make_organization_data("org_to_reactivate")
            )["id"]
        )

        # 1 query to load list of existing orgs,
        # 1 query to filter for only inactive existing orgs,
        # 1 query for activate-existing, and 1 query for create-new.
        with self.assertNumQueries(4):
            api.bulk_add_organizations([
                existing_org,
                existing_org,
                existing_org,
                self.make_organization_data("org_to_reactivate"),
                self.make_organization_data("new_org_1"),
                self.make_organization_data("new_org_2"),
                self.make_organization_data("new_org_3"),
                self.make_organization_data("new_org_4"),
                self.make_organization_data("new_org_5"),
                self.make_organization_data("new_org_6"),
                self.make_organization_data("new_org_7"),
                self.make_organization_data("new_org_8"),
                self.make_organization_data("new_org_9"),
                self.make_organization_data("new_org_9"),  # Redundant.
                self.make_organization_data("new_org_9"),  # Redundant.
            ])
        assert len(api.get_organizations()) == 11


class BulkAddOrganizationCoursesTestCase(utils.OrganizationsTestCaseBase):
    """
    Tests for `api.bulk_add_organization_courses`.
    """

    def test_validation_errors(self):
        """
        Test the `bulk_add_organization_courses` raises validation errors on bad input,
        and no organization-course linkages are created.
        """
        valid_org = self.make_organization_data("valid_org")
        invalid_org = {"description": "org with no short_name!"}
        valid_course_key = "course-v1:a+b+c"
        invalid_course_key = "NOT-A-COURSE-KEY"

        # Any bad org data or bad course key should cause the bulk-add to raise.
        with self.assertRaises(exceptions.InvalidCourseKeyException):
            api.bulk_add_organization_courses([
                (valid_org, valid_course_key),
                (valid_org, invalid_course_key),
            ])
        with self.assertRaises(exceptions.InvalidOrganizationException):
            api.bulk_add_organization_courses([
                (valid_org, valid_course_key),
                (invalid_org, valid_course_key),
            ])

        # In either case, no data should've been written for `valid_org`.
        assert len(api.get_organization_courses(valid_org)) == 0

    def test_add_no_organization_courses(self):
        """
        Test that `bulk_add_organization_courses` works given an an empty list.
        """
        # 1 query to load list of existing org-courses.
        with self.assertNumQueries(1):
            api.bulk_add_organization_courses([])

    def test_dry_run(self):
        """
        Test that `bulk_add_organization_courses` does nothing when `dry_run` is
        specified (except logging).
        """
        org_a = api.add_organization(self.make_organization_data("org_a"))
        course_key_x = CourseKey.from_string("course-v1:x+x+x")

        would_create, would_reactivate = api.bulk_add_organization_courses(
            [(org_a, course_key_x)], dry_run=True
        )

        assert api.get_organization_courses(org_a) == []
        assert would_create == {("org_a", "course-v1:x+x+x")}
        assert would_reactivate == set()

    def test_edge_cases(self):
        """
        Test that bulk_add_organization_courses handles a few edge cases as expected.
        """
        org_a = api.add_organization(self.make_organization_data("org_a"))
        org_b = api.add_organization(self.make_organization_data("org_b"))
        course_key_x = CourseKey.from_string("course-v1:x+x+x")
        course_key_y = CourseKey.from_string("course-v1:y+y+y")
        course_key_z = CourseKey.from_string("course-v1:z+z+z")

        # Add linkage A->X
        api.add_organization_course(org_a, course_key_x)

        # Add and then remove (under the hood: deactivate) linkage between A->Y.
        api.add_organization_course(org_a, course_key_y)
        api.remove_organization_course(org_a, course_key_y)

        # Add and then remove (under the hood: deactivate) linkage between A->Z.
        # This should NOT be reactivated, as we don't include it in the bulk_add call.
        api.add_organization_course(org_a, course_key_z)
        api.remove_organization_course(org_a, course_key_z)

        # 1 query to load list of existing linkages,
        # 1 query to fetch organizations for existing linkages,
        # 1 query to ensure all existing linkages active,
        # 1 query to get organiations for new linkages,
        # 1 query to create new linkages.
        with self.assertNumQueries(5):
            created, reactivated = api.bulk_add_organization_courses([

                # A->X: Existing linkage, should be a no-op.
                (org_a, course_key_x),

                # B->Y: Should create new linkage.
                (org_b, course_key_y),

                # A->Y: Is an inactive linkage; should be re-activated.
                (org_a, course_key_y),

                # B->Y: Is already in this list; shouldn't affect anything.
                (org_b, course_key_y),

                # B->Z: Adding with a stringified course id; should work as if we
                #       used the course key object.
                (org_b, str(course_key_z)),

                # B->Z: Adding again with the course key object; should be a no-op.
                (org_b, course_key_z),
            ])

        # Org A was linked to courses X and Y.
        # Org A also has an inactive link to course Z that we never re-activated.
        org_a_courses = api.get_organization_courses(org_a)
        assert {
            org_course["course_id"] for org_course in org_a_courses
        } == {
            "course-v1:x+x+x", "course-v1:y+y+y"
        }

        # Based on return value, make sure the expected breakdown of
        #    created vs. reactivated vs. not touched
        # is correct for the org-course linkages passed to `bulk_add_organization_courses`.
        assert created == {
            ("org_b", str(course_key_y)),
            ("org_b", str(course_key_z)),
        }
        assert reactivated == {
            ("org_a", str(course_key_y)),
        }

        # Org B was linked to courses Y and Z.
        org_b_courses = api.get_organization_courses(org_b)
        assert {
            org_course["course_id"] for org_course in org_b_courses
        } == {
            "course-v1:y+y+y", "course-v1:z+z+z"
        }

    def test_creation_of_inactive_organization_courses(self):
        """
        Test that the `activate` parameter can be set to `False`, which should
        cause new linkages to be created as "inactive" and stop organizations
        from being re-activated.
        """
        # Create one active and one inactive organization.
        active_org = api.add_organization(self.make_organization_data("active_org"))
        inactive_org = api.add_organization(self.make_organization_data("inactive_org"))
        api.remove_organization(inactive_org["id"])

        # Course X starts with existing active linkages to both orgs.
        course_key_x = CourseKey.from_string("course-v1:x+x+x")
        api.add_organization_course(active_org, course_key_x)
        api.add_organization_course(inactive_org, course_key_x)

        # Course Y starts with existing inactive linkages to both orgs.
        course_key_y = CourseKey.from_string("course-v1:y+y+y")
        api.add_organization_course(active_org, course_key_y)
        api.remove_organization_course(active_org, course_key_y)
        api.add_organization_course(inactive_org, course_key_y)
        api.remove_organization_course(inactive_org, course_key_y)

        # Course Z starts with no linkages.
        course_key_z = CourseKey.from_string("course-v1:z+z+z")

        # Bulk-add inactive linkages from every org to every course (6 linkages total).
        # This should be a no-op for existing linkages (whether active or inactive).
        linkages_to_bulk_add = list(
            itertools.product(
                (active_org, inactive_org),
                (course_key_x, course_key_y, course_key_z),
            )
        )
        created, reactivated = api.bulk_add_organization_courses(
            linkages_to_bulk_add, activate=False
        )

        # Based on the return value, make sure the expected breakdown of
        #    created vs. reactivated vs. not touched
        # is correct for the organizations passed to `bulk_add_organization_courses`.
        assert reactivated == set()
        assert created == {
            ("active_org", str(course_key_z)),
            ("inactive_org", str(course_key_z)),
        }

        # Check the final state of all existing linkages.
        assert models.OrganizationCourse.objects.count() == 6
        assert {
            (linkage.organization.short_name, str(linkage.course_id), linkage.active)
            for linkage in models.OrganizationCourse.objects.all()
        } == {
            ("active_org", "course-v1:x+x+x", True),
            ("inactive_org", "course-v1:x+x+x", True),
            ("active_org", "course-v1:y+y+y", False),
            ("inactive_org", "course-v1:y+y+y", False),
            ("active_org", "course-v1:z+z+z", False),
            ("inactive_org", "course-v1:z+z+z", False),
        }

    def test_add_several_organization_courses(self):
        """
        Test that the query_count of bulk_add_organization_courses does not increase
        when given more organization-course linkages to add.
        """
        org_a = api.add_organization(self.make_organization_data("org_a"))
        org_b = api.add_organization(self.make_organization_data("org_b"))
        org_c = api.add_organization(self.make_organization_data("org_c"))
        course_key_x = CourseKey.from_string("course-v1:x+x+x")
        course_key_y = CourseKey.from_string("course-v1:y+y+y")
        course_key_z = CourseKey.from_string("course-v1:z+z+z")

        # Add linkage A->X.
        api.add_organization_course(org_a, course_key_x)

        # Add linkage A->Y, and then remove (actually: deactivate).
        api.add_organization_course(org_a, course_key_y)
        api.remove_organization_course(org_a, course_key_y)

        # 1 query to load list of existing linkages,
        # 1 query to fetch organizations for existing linkages,
        # 1 query to ensure all existing linkages active,
        # 1 query to get organiations for new linkages,
        # 1 query to create new linkages.
        with self.assertNumQueries(5):
            api.bulk_add_organization_courses([
                (org_a, course_key_x),  # Already existing.
                (org_a, course_key_x),  # Already existing.
                (org_a, course_key_y),  # Reactivation.
                (org_a, course_key_z),  # The rest are new.
                (org_b, course_key_x),
                (org_b, course_key_y),
                (org_b, course_key_z),
                (org_c, course_key_x),
                (org_c, course_key_y),
                (org_c, course_key_z),
                (org_c, course_key_z),  # Redundant.
                (org_c, course_key_z),  # Redundant.
            ])
        assert len(api.get_organization_courses(org_a)) == 3
        assert len(api.get_organization_courses(org_b)) == 3
        assert len(api.get_organization_courses(org_c)) == 3



# Generated by Django 2.2.16 on 2020-11-09 16:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('organizations', '0001_squashed_0007_historicalorganization'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalorganization',
            name='short_name',
            field=models.CharField(db_index=True, help_text='Unique, short string identifier for organization. Please do not use spaces or special characters. Only allowed special characters are period (.), hyphen (-) and underscore (_).', max_length=255, verbose_name='Short Name'),
        ),
        migrations.AlterField(
            model_name='organization',
            name='short_name',
            field=models.CharField(help_text='Unique, short string identifier for organization. Please do not use spaces or special characters. Only allowed special characters are period (.), hyphen (-) and underscore (_).', max_length=255, unique=True, verbose_name='Short Name'),
        ),
    ]



"""
Permissions for organization viewsets.
"""
from rest_framework.permissions import BasePermission


class UserIsStaff(BasePermission):
    """
    Custom Permission class to check user is a publisher user or a staff user.
    """
    def has_permission(self, request, view):
        if request.method == 'PUT':
            return request.user.is_staff or request.user.is_superuser
        return True



""" Django admin pages for organization models """
from django.contrib import admin, messages
from django.utils.translation import gettext_lazy as _

from organizations.models import Organization, OrganizationCourse


class ActivateDeactivateAdminMixin:
    """
    Provides the activate_selected and deactivate_select bulk actions.

    Hides the delete_selected actions; we'd much rows are deactivated than
    deleted.
    """

    HISTORY_DISCLAIMER = _(
        "Please note: as a bulk action, this will not be reflected in the model's history table."
    )

    def get_actions(self, request):
        """ Return set of Django admin actions, removing the delete action """
        actions = super().get_actions(request)

        # Remove the delete action.
        if 'delete_selected' in actions:  # pragma: no cover
            del actions['delete_selected']

        return actions

    @admin.action(
        description=_('Activate selected entries')
    )
    def activate_selected(self, request, queryset):
        """ Activate the selected entries. """
        count = queryset.count()
        queryset.update(active=True)
        model_name = self.__class__.__name__

        if count == 1:
            message = _('1 {model_name} entry was successfully activated.')
        else:
            message = _('{count} {model_name} entries were successfully activated.')
        message = message.format(count=count, model_name=model_name)  # pylint: disable=no-member
        self.message_user(request, message)
        self.message_user(request, self.HISTORY_DISCLAIMER, level=messages.WARNING)

    @admin.action(
        description=_('Deactivate selected entries')
    )
    def deactivate_selected(self, request, queryset):
        """ Deactivate the selected entries. """
        count = queryset.count()
        queryset.update(active=False)
        model_name = self.__class__.__name__

        if count == 1:
            message = _('1 {model_name} entry was successfully deactivated.')
        else:
            message = _('{count} {model_name} entries were successfully deactivated.')
        message = message.format(count=count, model_name=model_name)  # pylint: disable=no-member
        self.message_user(request, message)
        self.message_user(request, self.HISTORY_DISCLAIMER, level=messages.WARNING)


@admin.register(Organization)
class OrganizationAdmin(ActivateDeactivateAdminMixin, admin.ModelAdmin):
    """ Admin for the Organization model. """
    actions = ['activate_selected', 'deactivate_selected']
    list_display = ('name', 'short_name', 'logo', 'active',)
    list_filter = ('active',)
    ordering = ('name', 'short_name',)
    readonly_fields = ('created',)
    search_fields = ('name', 'short_name',)


@admin.register(OrganizationCourse)
class OrganizationCourseAdmin(ActivateDeactivateAdminMixin, admin.ModelAdmin):
    """ Admin for the OrganizationCourse model. """
    actions = ['activate_selected', 'deactivate_selected']
    list_display = ('course_id', 'organization', 'active')
    list_filter = ('active',)
    ordering = ('course_id', 'organization__name',)
    search_fields = ('course_id', 'organization__name', 'organization__short_name',)

    def formfield_for_foreignkey(self, db_field, request=None, **kwargs):
        # Only display active Organizations.
        if db_field.name == 'organization':  # pragma: no branch
            kwargs['queryset'] = Organization.objects.filter(active=True).order_by('name')

        return super().formfield_for_foreignkey(db_field, request, **kwargs)



"""
Organizations Views Test Cases.
"""
import json
import ddt
from django.urls import reverse
from django.test import TestCase

from organizations.models import Organization
from organizations.serializers import OrganizationSerializer
from organizations.tests.factories import UserFactory, OrganizationFactory


@ddt.ddt
class TestOrganizationsView(TestCase):
    """ Test Organizations View."""

    def setUp(self):
        super().setUp()

        self.user_password = 'test'
        self.user = UserFactory(password=self.user_password, is_superuser=True)
        self.organization = OrganizationFactory.create()
        self.organization_list_url = reverse('v0:organization-list')
        self.client.login(username=self.user.username, password=self.user_password)

    def _get_organization_url(self, organization):
        """ Return organization specific URL."""
        return reverse('v0:organization-detail', kwargs={'short_name': organization.short_name})

    def test_authentication_required(self):
        """ Verify that authentication is required to access view."""
        self.client.logout()
        response = self.client.get(self.organization_list_url)
        self.assertEqual(response.status_code, 401)

    def test_authenticated_user(self):
        """ Verify that the authenticated user gets data."""
        OrganizationFactory.create()
        response = self.client.get(self.organization_list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['results']), 2)

    def test_single_organization(self):
        """verify single organization data could be fetched using short name"""
        url = self._get_organization_url(self.organization)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, OrganizationSerializer(self.organization).data)

    def test_inactive_organization(self):
        """ Verify inactive organization are filtered out."""
        organization = OrganizationFactory(active=False)
        url = self._get_organization_url(organization)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_nonexistent_organization(self):
        """ Verify that nonexistent organization have proper status code."""
        url = reverse('v0:organization-detail', kwargs={'short_name': 'dummy'})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    @ddt.data(
        # hyphen in short name
        {
            'name': 'example-name',
            'short_name': 'example-short-name',
            'description': 'example-description',
        },
        # period in short name
        {
            'name': 'example.name',
            'short_name': 'example.short.name',
            'description': 'example_description',
        },
        # underscore in short name
        {
            'name': 'example_name',
            'short_name': 'example_short_name',
            'description': 'example_description',
        }
    )
    def test_create_organization(self, org_data):
        """ Verify Organization can be created via PUT endpoint. """
        url = reverse('v0:organization-detail', kwargs={'short_name': org_data['short_name']})
        response = self.client.put(url, json.dumps(org_data), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], org_data['name'])
        self.assertEqual(response.data['short_name'], org_data['short_name'])
        self.assertEqual(response.data['description'], org_data['description'])
        orgs = Organization.objects.all()
        self.assertEqual(len(orgs), 2)

    def test_update_organization(self):
        """ Verify Organizations can be updated via PUT endpoint. """
        org = OrganizationFactory()
        data = {
            'name': 'changed-name',
            'short_name': org.short_name,
        }
        url = reverse('v0:organization-detail', kwargs={'short_name': org.short_name})
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], data['name'])
        self.assertEqual(response.data['short_name'], org.short_name)
        self.assertEqual(response.data['description'], org.description)
        orgs = Organization.objects.all()
        self.assertEqual(len(orgs), 2)

    def test_update_reactivates_inactive_organization(self):
        """ Verify inactive Organization can be updated via PUT endpoint, reactivating them. """
        org = OrganizationFactory(active=False)
        data = {
            'name': 'changed-name',
            'short_name': org.short_name,
        }
        url = reverse('v0:organization-detail', kwargs={'short_name': org.short_name})
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['active'], True)
        self.assertEqual(response.data['name'], data['name'])
        self.assertEqual(response.data['short_name'], org.short_name)
        self.assertEqual(response.data['description'], org.description)

    def test_activeness_may_not_be_specified(self):
        """ Verify that 'active' cannot be specified through the HTTP API. """
        org = OrganizationFactory(active=False)
        data = {
            'name': 'changed-name',
            'short_name': org.short_name,
            'active': False,
        }
        url = reverse('v0:organization-detail', kwargs={'short_name': org.short_name})
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 400)

    def test_patch_endpoint(self):
        """ Verify PATCH endpoint returns 405 because we use PUT for create and update"""
        url = reverse('v0:organization-detail', kwargs={'short_name': 'dummy'})
        response = self.client.patch(url, json={})
        self.assertEqual(response.status_code, 405)

    def test_post_endpoint(self):
        """ Verify POST endpoint returns 405 because we use PUT for create and update"""
        url = reverse('v0:organization-detail', kwargs={'short_name': 'dummy'})
        response = self.client.post(url, json={})
        self.assertEqual(response.status_code, 405)

    def test_delete_endpoint(self):
        """ Verify DELETE endpoint returns 405 because we don't support it"""
        url = reverse('v0:organization-detail', kwargs={'short_name': 'dummy'})
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 405)

    def test_create_as_only_staff_user(self):
        self.user.is_staff = True
        self.user.is_superuser = False
        self.user.save()

        data = {
            'name': 'example-name',
            'short_name': 'example-short-name',
            'description': 'example-description',
        }
        url = reverse('v0:organization-detail', kwargs={'short_name': data['short_name']})
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 200)

    def test_create_as_non_staff_and_non_admin_user(self):
        self.user.is_superuser = False
        self.user.save()

        data = {
            'name': 'example-name',
            'short_name': 'example-short-name',
            'description': 'example-description',
        }
        url = reverse('v0:organization-detail', kwargs={'short_name': data['short_name']})
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 403)



"""
URLS for organizations
"""
from django.urls import include, re_path

app_name = 'organizations'  # pylint: disable=invalid-name
urlpatterns = [
    re_path(r'^v0/', include('organizations.v0.urls')),
]



"""
Utility module for Organizations test cases
"""
from django.test import TestCase

from opaque_keys.edx.keys import CourseKey


class OrganizationsTestCaseBase(TestCase):
    """
    Parent/Base class for Organizations test cases
    """

    def setUp(self):
        """
        Helper method for test case scaffolding
        """
        super().setUp()
        self.test_course_key = CourseKey.from_string('the/course/key')

    @staticmethod
    def make_organization_data(short_name):
        """
        Make a fake organization dictionary, distinguished by a `short_name`.

        The `short_name` should be a valid short_name string, i.e. [0-9a-z_-].
        """
        return {
            'short_name': short_name,
            'name': f"Name of {short_name}",
            'description': f"Description of {short_name}",
        }



# pylint: disable=expression-not-assigned
"""
Application data management/abstraction layer.  Responsible for:

1) Accessing information from various resources:
* Internal application state  (through local models in models.py)
* External application state  (ORM bindings with other apps, yuck)
* Remote data services (through service adapters in resources.py)

2) Calculating derivative information from existing state:
* Algorithms and data manipulations
* Aggregations
* Annotations
* Alternative data representations

Accepts and returns standard Python data structures (dicts, arrays of dicts)
for easy consumption and manipulation by callers -- the queryset stops here!

When the time comes for remote resources, import the module like so:
if getattr(settings, 'TEST_MODE', False):
    import organizations.tests.mocks.resources as remote
else:
    import organizations.resources as remote
"""
import logging

from django.db.models.functions import Lower

from . import exceptions
from . import models as internal
from . import serializers


log = logging.getLogger(__name__)


# PRIVATE/INTERNAL METHODS (public methods located further down)
def _activate_record(record):
    """
    Enables database records by setting the 'active' attribute to True
    The queries in this module filter out inactive records as part of their criteria
    This effectively allows us to soft-delete records so they are not lost forever
    """
    record.active = True
    record.save()


def _inactivate_record(record):
    """
    Disables database records by setting the 'active' attribute to False
    The queries in this module filter out inactive records as part of their criteria
    This effectively allows us to soft-delete records so they are not lost forever
    """
    record.active = False
    record.save()


def _activate_organization(organization_id):
    """
    Activates an inactivated (soft-deleted) organization as well as any inactive relationships
    """
    [_activate_organization_course_relationship(record) for record
     in internal.OrganizationCourse.objects.filter(organization_id=organization_id, active=False)]

    [_activate_record(record) for record
     in internal.Organization.objects.filter(id=organization_id, active=False)]


def _inactivate_organization(organization_id):
    """
    Inactivates an activated organization as well as any active relationships
    """
    [_inactivate_organization_course_relationship(record) for record
     in internal.OrganizationCourse.objects.filter(organization_id=organization_id, active=True)]

    [_inactivate_record(record) for record
     in internal.Organization.objects.filter(id=organization_id, active=True)]


def _activate_organization_course_relationship(relationship):  # pylint: disable=invalid-name
    """
    Activates an inactive organization-course relationship
    """
    # If the relationship doesn't exist or the organization isn't active we'll want to raise an error
    relationship = internal.OrganizationCourse.objects.get(
        id=relationship.id,
        active=False,
        organization__active=True
    )
    _activate_record(relationship)


def _inactivate_organization_course_relationship(relationship):  # pylint: disable=invalid-name
    """
    Inactivates an active organization-course relationship
    """
    relationship = internal.OrganizationCourse.objects.get(
        id=relationship.id,
        active=True
    )
    _inactivate_record(relationship)


# PUBLIC METHODS
def create_organization(organization):
    """
    Inserts a new organization into app/local state given the following dictionary:
    {
        'short_name': string,
        'name': string,
        'description': string (optional),
        'logo': string (optional),
    }

    If an organization with the given `short_name` already exists, we will just
    activate that organization, but not update it.

    Returns an updated dictionary including a new 'id': integer field/value
    """
    if not (organization.get('name') and organization.get('short_name')):
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    organization_obj = serializers.deserialize_organization(organization)
    try:
        organization = internal.Organization.objects.get(
            short_name=organization_obj.short_name,
        )
        # If the organization exists, but was inactivated, we can simply turn it back on
        if not organization.active:
            _activate_organization(organization.id)
    except internal.Organization.DoesNotExist:
        organization = internal.Organization.objects.create(
            short_name=organization_obj.short_name,
            name=organization_obj.name,
            description=organization_obj.description,
            logo=organization_obj.logo,
            active=True
        )
    return serializers.serialize_organization(organization)


def bulk_create_organizations(organizations, dry_run=False, activate=True):
    """
    Efficiently insert multiple organizations into the database.

    Arguments:

        organizations (iterable[dict]):

            An iterable of `organization` dictionaries, each in the following format:
            {
                'short_name': string,
                'name': string,
                'description': string (optional),
                'logo': string (optional),
            }

            Organizations that do not already exist (by short_name) will be created.
            Organizations that already exist (by short_name) will be activated,
            but their name, description, and logo will be left as-is in the database.

            If multiple organizations share a `short_name`, the first organization
            in `organizations` will be used, and the latter ones ignored.

        dry_run (bool):
            Optional, defaulting to False.
            If True, don't apply changes, but still return organizations
            that would have been created or reactivated.

        activate (bool):
            Optional, defaulting to True.
            If True, missing organizations will be created with active=True,
            and existing-but-inactivate organizations will be reactivated.
            If False, missing organizations will be created with active=False,
            and existing-but-inactivate organizations will be left as inactivate.

    Returns: tuple[set[str], set[str]]

        A tuple in the form: (
            short names of organizations that were newly created,
            short names of organizations that we reactivated
        )
        If `activate` was supplied as False, the set of reactivated organizations
        will always be empty.
    """
    # Collect organizations by short name, dropping conflicts as necessary.
    organization_objs = [
        # This deserializes the dictionaries into Organization instances that
        # have not yet been saved to the db.
        serializers.deserialize_organization(organization_dict)
        for organization_dict in organizations
    ]
    # Make sure `active` is set correctly on the deserialized organizations.
    for organization_obj in organization_objs:
        organization_obj.active = activate
    organizations_by_short_name = {}
    for organization in organization_objs:
        # Make sure to lowercase short_name because MySQL UNIQUE is case-insensitive.
        short_name_lower = organization.short_name.lower()
        # Purposefully drop lowered short_names we've already seen, as noted in docstring.
        org_with_same_short_name = organizations_by_short_name.get(short_name_lower)
        if org_with_same_short_name:
            log.info(
                "Dropping organization from bulk_create batch, "
                "as an organization with the same short_name is already being "
                "created in this batch. Dropped data: %r. Kept data: %r.",
                serializers.serialize_organization(organization),
                serializers.serialize_organization(org_with_same_short_name),
            )
            continue
        organizations_by_short_name[short_name_lower] = organization

    # Find out which organizations we need to reactivate vs. create.
    existing_organizations = query_organizations_by_short_name(organizations_by_short_name)
    existing_organization_short_names = {
        short_name.lower()
        for short_name
        in existing_organizations.values_list("short_name", flat=True)
    }
    organizations_to_create = [
        organization
        for short_name, organization in organizations_by_short_name.items()
        if short_name.lower() not in existing_organization_short_names
    ]
    if activate:
        organizations_to_reactivate = existing_organizations.filter(active=False)
    else:
        organizations_to_reactivate = internal.Organization.objects.none()

    # Collect sets of orgs that will be reactivated and created,
    # so that we can have an informative return value.
    short_names_of_organizations_to_reactivate = set(
        organizations_to_reactivate.values_list("short_name", flat=True)
    )
    short_names_of_organizations_to_create = {
        org.short_name for org in organizations_to_create
    }

    # If not a dry run,
    # re-activate existing organizations, and create the new ones.
    # If `activate==False`, then `organizations_to_reactivate` will be empty.
    if not dry_run:
        organizations_to_reactivate.update(active=True)
        internal.Organization.objects.bulk_create(organizations_to_create)

    return (
        short_names_of_organizations_to_create,
        short_names_of_organizations_to_reactivate,
    )


def update_organization(organization):
    """
    Updates an existing organization in app/local state
    Returns a dictionary representation of the object
    """
    organization_obj = serializers.deserialize_organization(organization)
    try:
        organization = internal.Organization.objects.get(id=organization_obj.id)
        organization.name = organization_obj.name
        organization.short_name = organization_obj.short_name
        organization.description = organization_obj.description
        organization.logo = organization_obj.logo
        organization.active = organization_obj.active
    except internal.Organization.DoesNotExist:
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    return serializers.serialize_organization(organization)


def delete_organization(organization):
    """
    Inactivates an existing organization from app/local state
    No return currently defined for this operation
    """
    organization_obj = serializers.deserialize_organization(organization)
    _inactivate_organization(organization_obj.id)


def fetch_organization(organization_id):
    """
    Retrieves a specific organization from app/local state
    Returns a dictionary representation of the object
    """
    organization = {'id': organization_id}
    if not organization_id:
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    organizations = serializers.serialize_organizations(
        internal.Organization.objects.filter(id=organization_id, active=True)
    )
    if not organizations:
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    return organizations[0]


def fetch_organization_by_short_name(organization_short_name):
    """
    Retrieves a specific organization from app/local state by short name
    Returns a dictionary representation of the object
    """
    organization = {'short_name': organization_short_name}
    if not organization_short_name:
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    organizations = serializers.serialize_organizations(internal.Organization.objects.filter(
        active=True, short_name=organization_short_name
    ))
    if not organizations:
        exceptions.raise_exception("organization", organization, exceptions.InvalidOrganizationException)
    return organizations[0]


def fetch_organizations():
    """
    Retrieves the set of active organizations from app/local state
    Returns a list-of-dicts representation of the object
    """
    return serializers.serialize_organizations(internal.Organization.objects.filter(active=True))


def create_organization_course(organization, course_key):
    """
    Inserts a new organization-course relationship into app/local state
    No response currently defined for this operation
    """
    organization_obj = serializers.deserialize_organization(organization)
    try:
        relationship = internal.OrganizationCourse.objects.get(
            organization=organization_obj,
            course_id=str(course_key)
        )
        # If the relationship exists, but was inactivated, we can simply turn it back on
        if not relationship.active:
            _activate_organization_course_relationship(relationship)
    except internal.OrganizationCourse.DoesNotExist:
        relationship = internal.OrganizationCourse.objects.create(
            organization=organization_obj,
            course_id=str(course_key),
            active=True
        )


def bulk_create_organization_courses(
        organization_course_pairs,
        dry_run=False,
        activate=True,
):
    """
    Efficiently insert multiple organization-course relationships into the database.

    Arguments:

        organization_course_pairs (iterable[tuple[dict, CourseKey]]):

            An iterable of (organization_data, course_key) pairs.

            Organization-course linkages that DO NOT already exist will be created.
            Organization-course linkages that DO already exist will be activated, if inactive.

            Assumption: All provided organizations already exist in the DB.

        dry_run (bool):
            Optional, defaulting to False.
            If True, don't apply changes, but still return organization-course
            linkages that would have been created or reactivated.

        activate (bool):
            Optional, defaulting to True.
            If True, missing linkages will be created with active=True,
            and existing-but-inactive linkages will be reactivated.
            If False, missing linkages will be created with active=False,
            and existing-but-inactive linkages will be left as inactive.

    Returns: tuple[
                set[tuple[str, str]],
                set[tuple[str, str]]
             ]

        A tuple in the form: (
            organization-course linkages that we newly created,
            organization-course linkages that we reactivated
        )
        where an "organization-course" linkage is a tuple in the form:
            (organization short name, course key string).
        If `activate` was supplied as False, the set of reactivated linkages will
        always be empty.
    """
    def linkage_to_pair(linkage):
        """
        Given a `OrganizationCourse` (aka org-course linkage) instance,
        return a pair of strings:
        (lowercased organization short name, stringified course key)

        We normalize short_names to lowercase because MySQL UNIQUE is case-insensitive.
        """
        return linkage.organization.short_name.lower(), linkage.course_id

    # Grab all org-course linkages from db (whether active or inactive).
    db_linkages = list(
        internal.OrganizationCourse.objects.all().prefetch_related(
            'organization'
        )
    )

    # For the organizations that have been requested for creation,
    # build a set of (lowered org shortname, course key string) linkage pairs.
    # This will remove any duplicates.
    requested_linkage_pairs = {
        (organization_data["short_name"].lower(), str(course_key))
        for organization_data, course_key
        in organization_course_pairs
    }

    # Build the same set of pairs for linkages already in the db.
    db_linkage_pairs = {
        linkage_to_pair(linkage) for linkage in db_linkages
    }

    # The set of org-course linkages that we must CREATE
    # is the set of linkages that were REQUESTED
    # minus the set of linkages in the DATABASE.
    linkage_pairs_to_create = requested_linkage_pairs - db_linkage_pairs

    # The set of org-course linkages that we must ENSURE ARE ACTIVE
    # is the set of linkages that were REQUESTED
    # minus the set of linkages that WILL BE CREATED,
    # unless activate==False, in which case we won't ensure anything is active.
    if activate:
        linkage_pairs_to_ensure_active = requested_linkage_pairs - linkage_pairs_to_create
    else:
        linkage_pairs_to_ensure_active = set()

    # The linkages we must REACTIVATE
    # are those that we must ENSURE ARE ACTIVE
    # and are currently INACTIVE.
    linkages_to_reactivate = [
        linkage
        for linkage in db_linkages
        if (
            linkage_to_pair(linkage) in linkage_pairs_to_ensure_active
            and not linkage.active
        )
    ]

    # If this is a dry run, return before applying any changes to the db.
    linkage_pairs_to_reactivate = {
        linkage_to_pair(linkage)
        for linkage in linkages_to_reactivate
    }
    if dry_run:
        return linkage_pairs_to_create, linkage_pairs_to_reactivate

    # Bulk-reactivate existing organization-course linkages.
    # If `ids_of_linkages_to_reactivate` is an empty set, then this is a no-op.
    ids_of_linkages_to_reactivate = {linkage.id for linkage in linkages_to_reactivate}
    internal.OrganizationCourse.objects.filter(
        id__in=ids_of_linkages_to_reactivate
    ).update(
        active=True
    )

    # Load up a dict: (org short names, Organization model instances).
    # We need these Organizations so we can create linkages against them in the
    # next step.
    organizations_for_create = query_organizations_by_short_name(
        org for org, _ in linkage_pairs_to_create
    )
    organizations_for_create_by_short_name = {
        organization.short_name.lower(): organization
        for organization in organizations_for_create
    }

    # Bulk-create new organization-course linkages.
    internal.OrganizationCourse.objects.bulk_create([
        internal.OrganizationCourse(
            organization=organizations_for_create_by_short_name[org_short_name],
            course_id=course_id,
            active=activate,
        )
        for org_short_name, course_id
        in linkage_pairs_to_create
    ])
    return linkage_pairs_to_create, linkage_pairs_to_reactivate


def query_organizations_by_short_name(short_names):
    """
    Get a queryset of organizations from an iterable of organiztion short names.

    This is meant as a utility function for the bulk-create functions.
    It is not currently exposed through the API, but it could be.

    Canonicalize short names to lowercase due to case-insensitivity of
    MySQL UNIQUE constraint on `short_name`.

    Arguments:
        short_names (Iterable[str])

    Returns: QuerySet[Organization]
    """
    unique_short_names = {short_name.lower() for short_name in short_names}
    return internal.Organization.objects.annotate(
        short_name_lowered=Lower("short_name")
    ).filter(
        short_name_lowered__in=unique_short_names,
    )


def delete_organization_course(organization, course_key):
    """
    Removes an existing organization-course relationship from app/local state
    No response currently defined for this operation
    """
    try:
        relationship = internal.OrganizationCourse.objects.get(
            organization=organization['id'],
            course_id=str(course_key),
            active=True,
        )
        _inactivate_organization_course_relationship(relationship)
    except internal.OrganizationCourse.DoesNotExist:
        # If we're being asked to delete an organization-course link
        # that does not exist in the database then our work is done
        pass


def fetch_organization_courses(organization):
    """
    Retrieves the set of courses currently linked to the specified organization
    """
    organization_obj = serializers.deserialize_organization(organization)
    # Since django 5.2 unsaved objects can't be used in related queries so we use the pk directly
    queryset = internal.OrganizationCourse.objects.filter(
        organization_id=organization_obj.pk,
        active=True
    ).select_related('organization')
    return [serializers.serialize_organization_with_course(organization) for organization in queryset]


def fetch_course_organizations(course_key):
    """
    Retrieves the organizations linked to the specified course
    """
    queryset = internal.OrganizationCourse.objects.filter(
        course_id=str(course_key),
        active=True
    ).select_related('organization')
    return [serializers.serialize_organization_with_course(organization) for organization in queryset]


def delete_course_references(course_key):
    """
    Inactivates references to course keys within this app (ref: receivers.py and api.py)
    """
    [_inactivate_record(record) for record in internal.OrganizationCourse.objects.filter(
        course_id=str(course_key),
        active=True
    )]



"""
Add an organization through manage.py.
"""
import logging

from django.core.management import BaseCommand, CommandError

from organizations import api


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """Management command used to add an organization.

    Example: ./manage.py add_organization hogwarts "Hogwarts School of Witchcraft and Wizardry"
    """

    def add_arguments(self, parser):
        parser.add_argument('short_name')
        parser.add_argument('name')

    def handle(self, *args, **options):
        org_options = {
            key: options[key] for key in ['short_name', 'name']
        }

        if len(org_options['short_name']) > len(org_options['name']):
            fmt = (
                "The provided short_name ({short_name}) "
                "is longer than the provided name ({name}). "
                "You probably want to switch the order of the arguments."
            )
            raise CommandError(fmt.format(**org_options))

        created_org = api.add_organization(org_options)
        log_fmt = (
            "Created or activated organization: id={id}, "
            "short_name={short_name}, name='{name}'"
        )
        logger.info(log_fmt.format(**created_org))  # pylint: disable=logging-format-interpolation



"""
api.py is an interface module for Python-level integration with the
edx-organizations app.

In the current incarnation of this particular application, the API
operations are equivalent to the orchestration layer, which manages
the application's workflows.
"""
import logging

from django.conf import settings

from . import data
from . import exceptions
from . import validators


log = logging.getLogger(__name__)


# PRIVATE/INTERNAL FUNCTIONS

def _validate_course_key(course_key):
    """ Validation helper """
    if not validators.course_key_is_valid(course_key):
        exceptions.raise_exception(
            "CourseKey",
            course_key,
            exceptions.InvalidCourseKeyException
        )


def _validate_organization_data(organization_data):
    """ Validation helper """
    if not validators.organization_data_is_valid(organization_data):
        exceptions.raise_exception(
            "Organization",
            organization_data,
            exceptions.InvalidOrganizationException
        )


# PUBLIC FUNCTIONS
def add_organization(organization_data):
    """
    Passes a new organization to the data layer for storage
    """
    _validate_organization_data(organization_data)
    organization = data.create_organization(organization_data)
    return organization


def bulk_add_organizations(organization_data_items, dry_run=False, activate=True):
    """
    Efficiently store multiple organizations.

    Note: No `pre_save` or `post_save` signals for `Organization` will be
    triggered. This is due to the underlying Django implementation of `bulk_create`:
    https://docs.djangoproject.com/en/2.2/ref/models/querysets/#bulk-create

    Arguments:

        organizations (iterable[dict]):

            An iterable of `organization` dictionaries, each in the following format:
            {
                'short_name': string,
                'name': string,
                'description': string (optional),
                'logo': string (optional),
            }

            Organizations that do not already exist (by short_name) will be created.
            Organizations that already exist (by short_name) will be activated,
            but their name, description, and logo will be left as-is in the database.

            If multiple organizations share a `short_name`, the first organization
            in `organization_data_items` will be used, and the latter ones ignored.

        dry_run (bool):
            Optional, defaulting to False.
            If True, don't apply changes, but still return organizations
            that would have been created or reactivated.

        activate (bool):
            Optional, defaulting to True.
            If True, missing organizations will be created with active=True,
            and existing-but-inactive organizations will be reactivated.
            If False, missing organizations will be created with active=False,
            and existing-but-inactive organizations will be left as inactive.

    Raises:
        InvalidOrganizationException: One or more organization dictionaries
            have missing or invalid data; no organizations were created.

    Returns: tuple[set[str], set[str]]

        A tuple in the form: (
            short names of organizations that were newly created,
            short names of organizations that we reactivated
        )
        If `activate` was supplied as False, the set of reactivated linkages will
        always be empty.

        From an API layer point of view, the organizations that were "added"
        is the union of the organizations that were *newly created* and those
        that were *reactivated*. We distinguish between them in the return
        value to allow for richer reporting by users of this function.
    """
    for organization_data in organization_data_items:
        _validate_organization_data(organization_data)
        if "short_name" not in organization_data:
            raise exceptions.InvalidOrganizationException(
                f"Organization is missing short_name: {organization_data}"
            )
    return data.bulk_create_organizations(
        organization_data_items, dry_run=dry_run, activate=activate
    )


def edit_organization(organization_data):
    """
    Passes an updated organization to the data layer for storage
    """
    _validate_organization_data(organization_data)
    return data.update_organization(organization_data)


def get_organization(organization_id):
    """
    Retrieves the specified organization
    """
    return data.fetch_organization(organization_id)


def get_organization_by_short_name(organization_short_name):
    """
    Retrieves the organization filtered by short name
    """
    return data.fetch_organization_by_short_name(organization_short_name)


def get_organizations():
    """
    Retrieves the active organizations managed by the system
    """
    return data.fetch_organizations()


def remove_organization(organization_id):
    """
    Removes the specified organization
    """
    organization = {
        'id': organization_id,
    }
    data.delete_organization(organization)


def add_organization_course(organization_data, course_key):
    """
    Adds a organization-course link to the system
    """
    _validate_course_key(course_key)
    _validate_organization_data(organization_data)
    data.create_organization_course(
        organization=organization_data,
        course_key=course_key
    )


def bulk_add_organization_courses(
        organization_course_pairs,
        dry_run=False,
        activate=True,
):
    """
    Efficiently store multiple organization-course relationships.

    Note: No `pre_save` or `post_save` signals for `OrganizationCourse` will be
    triggered. This is due to the underlying Django implementation of `bulk_create`:
    https://docs.djangoproject.com/en/2.2/ref/models/querysets/#bulk-create

    Arguments:

        organization_course_pairs (iterable[tuple[dict, CourseKey]]):

            An iterable of (organization_data, course_key) pairs.

            We will ensure that these organization-course linkages exist.

            Assumption: All provided organizations already exist in storage.

        dry_run (bool):
            Optional, defaulting to False.
            If True, don't apply changes, but still return organization-course
            linkages that would have been created or reactivated.

        activate (bool):
            Optional, defaulting to True.
            If True, missing linkages will be created with active=True,
            and existing-but-inactive linkages will be reactivated.
            If False, missing linkages will be created with active=False,
            and existing-but-inactive linkages will be left as inactive.

    Raises:
        InvalidOrganizationException: One or more organization dictionaries
            have missing or invalid data.
        InvalidCourseKeyException: One or more course keys could not be parsed.
        (in case of either exception, no org-course linkages are created).

    Returns: tuple[
                set[tuple[str, CourseKey],
                set[tuple[str, CourseKey]
            ]

        A tuple in the form: (
            organization-course linkages that we newly created,
            organization-course linkages that we reactivated
        )
        where the `str` objects are organization short names.
        If `activate` was supplied as False, the set of reactivated linkages will
        always be empty.

        From an API layer point of view, the organization-
        course linkages that were "added" is the union of the linkages that were
        *newly created* and those that were *reactivated*.
        We distinguish between them in the return value to allow for richer
        reporting by users of this function.
    """
    for organization_data, course_key in organization_course_pairs:
        _validate_organization_data(organization_data)
        if "short_name" not in organization_data:
            raise exceptions.InvalidOrganizationException(
                f"Organization is missing short_name: {organization_data}"
            )
        _validate_course_key(course_key)
    return data.bulk_create_organization_courses(
        organization_course_pairs, dry_run=dry_run, activate=activate
    )


def get_organization_courses(organization_data):
    """
    Retrieves the set of courses for a given organization
    Returns an array of course identifiers
    """
    _validate_organization_data(organization_data)
    return data.fetch_organization_courses(organization=organization_data)


def remove_organization_course(organization, course_key):
    """
    Removes the specfied course from the specified organization
    """
    _validate_organization_data(organization)
    _validate_course_key(course_key)
    return data.delete_organization_course(course_key=course_key, organization=organization)


def get_course_organizations(course_key):
    """
    Retrieves the set of organizations for a given course
    Returns an array of dicts containing organizations
    """
    _validate_course_key(course_key)
    return data.fetch_course_organizations(course_key=course_key)


def get_course_organization(course_key):
    """
    Returns the first organization linked to a given course,
    or None if the course is not linked to any organizations.
    """
    _validate_course_key(course_key)
    course_organization = get_course_organizations(course_key)
    if course_organization:
        return course_organization[0]
    return None


def get_course_organization_id(course_key):
    """
    Returns the id of the first organization linked to a given course,
    or None if the course is not linked to any organizations.
    """
    course_org = get_course_organization(course_key)
    return course_org["id"] if course_org else None


def remove_course_references(course_key):
    """
    Removes course references from application state
    See edx-platform/lms/djangoapps/courseware/management/commands/delete_course_references.py
    """
    _validate_course_key(course_key)
    data.delete_course_references(course_key)


def ensure_organization(organization_short_name):
    """
    Ensure that an organization with the given short name exists.

    If the organization exists, then return it.
    If the organization does not exist, then:
        (a) If auto-create is enabled, create & return a new organization.
        (b) If auto-create is disabled, raise an InvalidOrganizationException.

    Arguments:
        organization_short_name (str)

    Returns: dict
        dictionary containing organization data

    Raises:
        InvalidOrganizationException (only if auto-create is disabled)
    """
    try:
        return get_organization_by_short_name(organization_short_name)
    except exceptions.InvalidOrganizationException:
        if not is_autocreate_enabled():
            raise
    log.info("Automatically creating new organization '%s'.", organization_short_name)
    return add_organization({
        "short_name": organization_short_name,
        "name": organization_short_name,
    })


def is_autocreate_enabled():
    """
    Return whether automatic organization creation is enabled.

    If True (default), calls to ``ensure_organization`` will auto-create organizations
    that do not already exist in the database.

    If False, calls to ``ensure_organization`` will fail for organizations that are not
    already in the database.
    """
    try:
        return bool(settings.ORGANIZATIONS_AUTOCREATE)
    except AttributeError:
        return True



"""
Data layer serialization operations.  Converts querysets to simple
python containers (mainly arrays and dicts).
"""
import requests

from django.core.files.base import ContentFile
from rest_framework import serializers

from organizations import models


class OrganizationSerializer(serializers.ModelSerializer):
    """ Serializes the Organization object."""
    logo_url = serializers.CharField(write_only=True, required=False)

    class Meta:
        model = models.Organization
        fields = ('id', 'created', 'modified', 'name', 'short_name', 'description', 'logo',
                  'active', 'logo_url',)

    def update_logo(self, obj, logo_url):
        if logo_url:  # pragma: no cover
            logo = requests.get(logo_url)  # pylint: disable=missing-timeout
            obj.logo.save(logo_url.split('/')[-1], ContentFile(logo.content))

    def create(self, validated_data):
        """
        New organizations created through the API are always Active
        """
        validated_data.update({'active': True})
        logo_url = validated_data.pop('logo_url', None)
        obj = super().create(validated_data)
        self.update_logo(obj, logo_url)
        return obj

    def update(self, instance, validated_data):
        """
        Existing organizations updated through the API always end up Active,
        regardless of whether or not they were previously active
        """
        validated_data.update({'active': True})
        logo_url = validated_data.pop('logo_url', None)
        super().update(instance, validated_data)
        self.update_logo(instance, logo_url)
        return instance


def serialize_organization(organization):
    """
    Organization object-to-dict serialization
    """
    return {
        'id': organization.id,
        'name': organization.name,
        'short_name': organization.short_name,
        'description': organization.description,
        'logo': organization.logo
    }


def serialize_organization_with_course(organization_course):
    """
    OrganizationCourse serialization (composite object)
    """
    return {
        'id': organization_course.organization.id,
        'name': organization_course.organization.name,
        'short_name': organization_course.organization.short_name,
        'description': organization_course.organization.description,
        'logo': organization_course.organization.logo,
        'course_id': organization_course.course_id
    }


def serialize_organizations(organizations):
    """
    Organization serialization
    Converts list of objects to list of dicts
    """
    return [serialize_organization(organization) for organization in organizations]


def deserialize_organization(organization_dict):
    """
    Organization dict-to-object serialization
    """
    return models.Organization(
        id=organization_dict.get('id'),
        name=organization_dict.get('name', ''),
        short_name=organization_dict.get('short_name', ''),
        description=organization_dict.get('description', ''),
        logo=organization_dict.get('logo', '')
    )



"""
Views for organizations end points.
"""
from django.http import Http404
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import mixins
from rest_framework import status
from rest_framework import viewsets
from rest_framework.authentication import SessionAuthentication
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from organizations.models import Organization
from organizations.permissions import UserIsStaff
from organizations.serializers import OrganizationSerializer


class OrganizationsViewSet(mixins.UpdateModelMixin, viewsets.ReadOnlyModelViewSet):
    """
    Organization view to:
        - list organization data (GET .../)
        - retrieve single organization (GET .../<short_name>)
        - create or update an organization via the PUT endpoint (PUT .../<short_name>)
    """
    queryset = Organization.objects.all()
    serializer_class = OrganizationSerializer
    # Include everything not a '+' - identical to discovery's COURSE_ID_REGEX.
    lookup_value_regex = '[^/+]+'
    lookup_field = 'short_name'
    authentication_classes = (JwtAuthentication, SessionAuthentication)
    permission_classes = (IsAuthenticated, UserIsStaff)

    def get_queryset(self):
        """
        Get the queryset to use in the request.

        For listing and retieving organizations, we only want to include
        active ones.

        For creating and updating organizations, we want to include all of
        them, which allows API users to "create" (i.e., reactivate)
        organizations that exist internally but are inactive.
        """
        if self.request.method == "GET":
            return self.queryset.filter(active=True)
        return self.queryset

    def update(self, request, *args, **kwargs):
        """
        We perform both Update and Create action via the PUT method.

        The 'active' field may not be specified via the HTTP API, since it
        is always assumed to be True. So:
            (1) new organizations created through the API are always Active, and
            (2) existing organizations updated through the API always end up Active,
                regardless of whether or not they were previously active.
        """
        if 'active' in self.request.data:
            raise ValidationError(
                "Value of 'active' may not be specified via Organizations HTTP API."
            )
        try:
            return super().update(request, *args, **kwargs)
        except Http404:
            serializer = OrganizationSerializer(data=request.data)
            serializer.is_valid(raise_exception=True)
            serializer.save()
            return Response(serializer.data)

    def partial_update(self, request, *args, **kwargs):
        """
        We disable PATCH because all updates and creates should use the PUT action above.
        """
        return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
